'''
Mass Extinction in Evolutionary Algorithms
Author: Kaelan Engholdt
Version: 10/07/2020

Mass extinction class for use in evolutionary algorithms, executes extinction events and methods of repopulation.
Users may save extinction parameter sets by adding additional 'if' code blocks to the '_saved_parameters' method.
Adjustable parameters are found in 'ext_params.py', consult the README.txt for more information.

Note: Due to the problem-specific nature of evolutionary algorithms, minor parts of this class must be finished by
      the user.

'''

# TODO user must copy this file and place it in the same directory as their evolutionary algorithm

import sys
from random import uniform
from math import log
import ext_params as ep
import ext_types as et


class Extinction:
    def __init__(self):
        pass
    
    # main method, handles top level functionality of the Extinction class
    def main(self, population, user_params, logpath, saved_idx=0):
        '''
        Parameters
        ----------
        :param population : the sorted parent population
            :type : list
        
        :param user_params : holds the relevant values needed from the user's evolutionary algorithm
            :type : list of the form [current_gen, max_gens, xo_prob, mut_prob, tourn_size]
        
        :param logpath : the absolute path to the log file, where information for the current run is stored
            :type : str
        
        :param saved_idx : the index into the set of extinction parameters saved in the '_saved_parameters' method
            :type : int
        
        Return Value
        ----------
        :return : the altered population, must be resorted by the user after it is returned to their algorithm
            :type : list
        '''
        
        # define population size
        population_size = len(population)
        
        # define user parameters
        generation = user_params[0]             # the current generation
        maximum_generations = user_params[1]    # the total generations the evolutionary algorithm will run for
        xo_prob = user_params[2]                # the probability of crossover occurring
        mut_prob = user_params[3]               # the probability of mutation occurring
        tourn_size = user_params[4]             # the tournament size
        
        # set the parameters to be returned back to the evolutionary algorithm
        ret_params = [xo_prob, mut_prob, tourn_size]
        
        # reset output string
        ep.ext_out_str = ""
        
        # if generations are zero-indexed, indicate as such
        if generation == 0:
            # indicate this evolutionary algorithm uses zero-indexed generations
            ep.ext_gens_zeroed = True
        
        # if generations are zero-indexed, remedy this issue
        if ep.ext_gens_zeroed:
            # increase generation count by 1 to avoid screwing up calculations
            generation += 1
        
        # perform extinction start-up operations upon first iteration
        if generation == 1 and ep.extinction_on:
            # change extinction parameters according to saved parameter index (uses ext_params.py if no index given)
            self._saved_parameters(saved_idx)
            
            # check if input values found in ext_params.py are valid
            self._error_checking(population_size, maximum_generations)
            
            # fetch all starting parameters for the log file
            self._fetch_params()
            
            # reset starting variables
            ep.ext_original_size = 0
            ep.ext_pop_dynamics = []
            ep.ext_occurred = False
            ep.ext_in_progress = False
            ep.ext_gens_tracker = 1
            ep.ext_generations = 0
            ep.ext_gens_zeroed = False
            ep.ext_values = []
            ep.ext_preserve = 0
            ep.ext_terminate = 0
            ep.i_ext_gens = 0
            ep.i_ext_hits = False
            ep.p_ext_prob = ep.p_ext_base_prob
            ep.repop_elite_pop = []
            ep.repop_in_progress = False
            ep.repop_gens_tracker = 1
            ep.newborn_count = 0
            
            # if population dynamics are not enabled
            if not ep.ext_dynamics:
                # set population dynamics strength to zero
                ep.ext_dyn_strength = 0.00
        
        # set variables to indicate an extinction event has not occurred yet this generation
        ep.ext_occurred = False
        ep.i_ext_hits = False
        
        # reset newborn count
        ep.newborn_count = 0
        
        # use newborn parameters, if enabled
        if ep.newborn_params_on and ep.extinction_on:
            # if the evolutionary algorithm uses the DEAP library
            if ep.ext_uses_DEAP:
                # check each member in the population for newborn status
                for member in population:
                    # if the member is a newborn, alter newborn parameters
                    if member[1]:
                        # increase number of generations this member has been a newborn by 1
                        member[2] += 1
                        
                        # check if the newborn member is old enough to become a normal member
                        if member[2] > ep.newborn_expires:
                            # switch the member from a newborn to a normal member
                            member[1] = False
                    
                    # if the member is still a newborn, increase newborn count
                    if member[1]:
                        ep.newborn_count += 1
            
            # if the evolutionary algorithm does not use the DEAP library
            else:
                # check each member in the population for newborn status
                for member in population:
                    # if the member is a newborn, alter newborn parameters
                    if member.ext_is_newborn:
                        # increase number of generations this member has been a newborn by 1
                        member.ext_newborn_gens += 1
                        
                        # check if the newborn member is old enough to become a normal member
                        if member.ext_newborn_gens > ep.newborn_expires:
                            # switch the member from a newborn to a normal member
                            member.ext_is_newborn = False
                    
                    # if the member is still a newborn, increase newborn count
                    if member.ext_is_newborn:
                        ep.newborn_count += 1
        
        # check if extinction will end, if enabled
        if ep.ext_end_extinction and ep.extinction_on:
            # check if the current generation has passed the ext_end_percent generation point
            if generation > int(ep.ext_end_percent * maximum_generations):
                # turn extinction off
                ep.extinction_on = False
        
        # proceed with extinction, ensure repopulation not in progress, ensure an extinction event is not in progress
        if ep.extinction_on and not ep.repop_in_progress and not ep.ext_in_progress:
            # uses interval extinction, if enabled
            if ep.interval_extinction_on:
                # call interval extinction method
                self._interval_extinction(generation)
            
            # uses probabilistic extinction, if enabled, and an extinction event hasn't occurred yet
            if ep.prob_extinction_on and not ep.ext_occurred:
                # check safeguard, proceed if sufficient generations have passed since last extinction event
                if ep.ext_generations >= ep.ext_safeguard:
                    # if using interval extinction in conjunction, check if the interval lands on this generation
                    if ep.interval_extinction_on and ep.i_ext_hits:
                        # call probabilistic extinction method
                        self._prob_extinction(generation)
                    
                    # if not using interval extinction in conjunction, proceed with probabilistic extinction
                    elif not ep.interval_extinction_on:
                        # call probabilistic extinction method
                        self._prob_extinction(generation)
            
            # uses fitness extinction, if enabled, and an extinction event hasn't occurred yet
            if ep.fit_extinction_on and not ep.ext_occurred:
                # check safeguard, proceed if sufficient generations have passed since last extinction event
                if ep.ext_generations >= ep.ext_safeguard:
                    # if using interval extinction in conjunction, check if the interval lands on this generation
                    if ep.interval_extinction_on and ep.i_ext_hits:
                        # call fitness extinction method
                        self._fit_extinction(population, population_size, generation)
                    
                    # if not using interval extinction in conjunction, proceed with fitness extinction
                    elif not ep.interval_extinction_on:
                        # call fitness extinction method
                        self._fit_extinction(population, population_size, generation)
        
        # if an extinction event occurred this generation
        if ep.ext_occurred:
            # number of generations since last extinction is zero
            ep.ext_generations = 0
            
            # define original population size before the extinction event occurred
            ep.ext_original_size = population_size
            
            # use elitism, if enabled
            if ep.ext_use_elitism:
                # determine percent of members to be kept
                ep.ext_preserve = int(ep.ext_elite_percent * population_size)
                
                # must keep at least one member
                if ep.ext_preserve < 1:
                    ep.ext_preserve = 1
                
                # if elitism is used in repopulation
                if ep.repop_use_elites:
                    # clear all old elite members
                    del ep.repop_elite_pop[:]
                    
                    # if elitism targets members at the front
                    if ep.ext_elite_target:
                        # add elite members to their own population
                        ep.repop_elite_pop.extend(population[:ep.ext_preserve])
                        
                        # temporarily remove the elite members from the parent population
                    
                    # if elitism targets members at the end
                    else:
                        # add elite members to their own population
                        ep.repop_elite_pop.extend(population[population_size - ep.ext_preserve :])
            
            # if elitism isn't in use
            else:
                ep.ext_preserve = 0
            
            # use anti-elitism, if enabled, for the first generation of extinction
            if ep.ext_use_anti_elite and ep.ext_anti_percent != 0 and ep.ext_gens_tracker == 1:
                # determine percent of members to be terminated
                ep.ext_terminate = int(ep.ext_anti_percent * population_size)
                
                # must terminate at least one member
                if ep.ext_terminate < 1:
                    ep.ext_terminate = 1
                
                # if anti-elitism targets members at the front
                if ep.ext_anti_target:
                    # kill all members to be terminated
                    del population[:ep.ext_terminate]
                
                # if anti-elitism targets members at the end
                else:
                    # kill all members to be terminated
                    del population[population_size - ep.ext_terminate:]
                
                # log anti-elitism deaths
                ep.ext_out_str += "Anti-Elitism Deaths: {}\n".format(ep.ext_terminate)
            
            # if anti-elitism isn't in use
            else:
                ep.ext_terminate = 0
            
            # indicate that an extinction event is now in progress
            ep.ext_in_progress = True
        
        # if an extinction event did not occur
        else:
            # increment the number of generations since the advent of the last extinction event
            ep.ext_generations += 1
        
        # if an extinction event is currently in progress, decrease population according to extinction parameters
        if ep.ext_in_progress:
            # call the extinction method
            population, ret_params = self._extinction(population, generation, ret_params)
        
        # if the population is repopulating, increase population according to repopulation parameters
        if ep.repop_in_progress:
            # call repopulation method
            population, ret_params = self._repopulate(population, generation, ret_params)
            
            # reset the number of generations since the previous repopulation finished
            ep.i_ext_gens = 0
        
        # if the population has finished repopulating
        else:
            # increment the number of generations since repopulation finished
            ep.i_ext_gens += 1
        
        # if an extinction event is currently in progress or the population is repopulating
        if ep.ext_in_progress or ep.repop_in_progress:
            # end the output string
            ep.ext_out_str += "\n"
        
        # if print statements are enabled, print all information from this generation to the console
        if ep.ext_print_on and ep.ext_out_str != "":
            print(ep.ext_out_str)
        
        # open the log file and save all information from this generation
        log_file = open(logpath, 'a')
        log_file.write(ep.ext_out_str)
        log_file.close()
        
        # return the altered population (unsorted and unevaluated) and the altered evolutionary algorithm parameters
        return population, ret_params[0], ret_params[1], ret_params[2]
    
    
    # interval extinction method
    def _interval_extinction(self, generation):
        '''
        Parameters
        ----------
        :param generation : the current generation
            :type : int
        '''
        
        # check if the extinction interval lands on this generation
        if (ep.i_ext_interval == 1) or (ep.i_ext_gens % (ep.i_ext_interval - 1) == 0 and ep.i_ext_gens != 0):
            # extinction occurs if i_ext_occurs is enabled
            if ep.i_ext_occurs:
                ep.ext_occurred = True
                
                # log extinction event
                ep.ext_out_str += "\n----------\n"
                ep.ext_out_str += "Interval Extinction Occurred: Generation {}\n".format(generation)
            
            # if i_ext_occurs is not enabled, indicate that the interval lands on this generation
            else:
                ep.i_ext_hits = True
    
    
    # probabilistic extinction method
    def _prob_extinction(self, generation):
        '''
        Parameters
        ----------
        :param generation : the current generation
            :type : int
        '''
        
        # determine if in extinction event occurs based on the current probability
        if uniform(0, 1) < ep.p_ext_prob:
            ep.ext_occurred = True
        
        # if an extinction event occurs, reset to base probability
        if ep.ext_occurred:
            # reset the current probability back to the base probability
            ep.p_ext_prob = ep.p_ext_base_prob
            
            # log extinction event
            ep.ext_out_str += "\n----------\n"
            ep.ext_out_str += "Probabilistic Extinction Occurred: Generation {}\n".format(generation)
        
        # if an extinction event did not occur, increase probability of extinction occurring next generation
        else:
            ep.p_ext_prob += ep.p_ext_inc
    
    
    # fitness extinction method
    def _fit_extinction(self, population, population_size, generation):
        '''
        Parameters
        ----------
        :param population : the sorted parent population
            :type : list
        
        :param population_size : the current size of the population
            :type : int
        
        :param generation : the current generation
            :type : int
        '''
        
        # if threshold extinction is in use
        if ep.f_ext_use_threshold:
            # determine number of members required to pass the higher fitness threshold
            fit_members = int(ep.f_ext_fit_percent_high * population_size)
            
            # define number if members in the parent population that pass the higher fitness threshold
            pop_fit = 0
            
            # TODO user must finish this code chunk, code instructions are found below
            '''
            1) Set 'pop_fit' equal to the number of members in the parent population that pass 'f_ext_threshold_high'
            '''
            
            # if there are sufficient fit members, extinction occurs
            if pop_fit >= fit_members:
                ep.ext_occurred = True
                
                # log extinction event
                ep.ext_out_str += "\n----------\n"
                ep.ext_out_str += "Fitness Threshold Extinction Occurred: Generation {}\n".format(generation)
        
        # if stagnation extinction is in use and an extinction event has not occurred yet
        if ep.f_ext_use_stagnation and not ep.ext_occurred:
            # determine number of members required to pass the lower fitness threshold
            fit_members = int(ep.f_ext_fit_percent_low * population_size)
            
            # define number of members in the parent population that pass the lower fitness threshold
            pop_fit = 0
            
            # TODO user must finish this code chunk, code instructions are found below
            '''
            1) Set 'pop_fit' equal to the number of members in the parent population that pass 'f_ext_threshold_low'
            '''
            
            # if sufficient members have not passed the lower fitness threshold, extinction occurs
            if not pop_fit >= fit_members:
                ep.ext_occurred = True
                
                # log extinction event
                ep.ext_out_str += "\n----------\n"
                ep.ext_out_str += "Fitness Stagnation Extinction Occurred: Generation {}\n".format(generation)
    
    
    # extinction method, decreases population according to extinction parameters
    def _extinction(self, population, generation, ret_params):
        '''
        Parameters
        ----------
        :param population : the sorted parent population
            :type : list
        
        :param generation : the current generation
            :type : int
        
        :param ret_params : the evolutionary algorithm parameters to be altered by an extinction event
            :type : list of the form [xo_prob, mut_prob, tourn_size]
        
        Return Value
        ----------
        :return : the altered population, decreased in size according to extinction parameters
            :type : list
        
        :return : the evolutionary algorithm parameters altered by the extinction event
            :type : list of the form [xo_prob, mut_prob, tourn_size]
        '''
        
        # set the current population size
        population_size = len(population)
        
        # determine number of members to be killed this generation
        del_num = self._calculate(ep.ext_original_size, population_size, ep.ext_duration, ep.ext_kills_constant)
        
        # determine number of members to be repopulated by repopulation dynamics this generation
        repopulate = ep.ext_pop_dynamics.pop(0)
        
        # determine the original number of members to be killed before population dynamics
        original_num = del_num - repopulate
        
        # redetermine number of members to be killed if newborn protection is in use
        if ep.newborn_params_on and ep.newborn_protection_on:
            # ensure del_num does not exceed the number of members to be preserved/protected
            if del_num > (population_size - (ep.ext_preserve + ep.newborn_count)):
                # set del_num to maximum possible number of members to be killed
                del_num = population_size - (ep.ext_preserve + ep.newborn_count)
        
        # redetermine number of members to be killed if newborn protection isn't in use
        else:
            # ensure del_num does not exceed the number of members to be preserved
            if del_num > (population_size - ep.ext_preserve):
                # set del_num to maximum possible number of members to be killed
                del_num = population_size - ep.ext_preserve
        
        # if del_num has fallen below zero due to double counting
        if del_num < 0:
            # no members will be killed this generation due to elitism and/or newborn protection restrictions
            del_num = 0
        
        # if alternate parameters are in use
        if ep.alt_params_on:
            # call the method to slowly increase alternate parameters to their heightened values
            ret_params = self._alt_param_increase(original_num, ret_params)
        
        # if elitism is in use and elitism targets members at the front
        if ep.ext_use_elitism and ep.ext_elite_target:
            # randomly kill members for extinction
            for i in range(del_num):
                # define loop variable
                invalid = True
                
                # keep choosing victims until the chosen victim is not a protected newborn
                while (invalid):
                    # calculate victim
                    victim = int(uniform(ep.ext_preserve, population_size - i))
                    
                    # if the evolutionary algorithm uses the DEAP library
                    if ep.ext_uses_DEAP:
                        # ensure the victim is not a protected newborn
                        if not (ep.newborn_params_on and ep.newborn_protection_on and population[victim][1]):
                            # kill the victim
                            del population[victim]
                            
                            # terminate while loop
                            invalid = False
                    
                    # if the evolutionary algorithm does not use the DEAP library
                    else:
                        # ensure the victim is not a protected newborn
                        if not (ep.newborn_params_on and ep.newborn_protection_on and
                                population[victim].ext_is_newborn):
                            
                            # kill the victim
                            del population[victim]
                            
                            # terminate while loop
                            invalid = False
        
        # if elitism is not in use or if elitism targets members at the end
        else:
            # randomly kill members for extinction
            for i in range(del_num):
                # define loop variable
                invalid = True
                
                # keep choosing victims until the chosen victim is not a protected newborn
                while (invalid):
                    # calculate victim
                    victim = int(uniform(0, population_size - (i + ep.ext_preserve)))
                    
                    # if the evolutionary algorithm uses the DEAP library
                    if ep.ext_uses_DEAP:
                        # ensure the victim is not a protected newborn
                        if not (ep.newborn_params_on and ep.newborn_protection_on and population[victim][1]):
                            # kill the victim
                            del population[victim]
                            
                            # terminate while loop
                            invalid = False
                    
                    # if the evolutionary algorithm does not use the DEAP library
                    else:
                        # ensure the victim is not a protected newborn
                        if not (ep.newborn_params_on and ep.newborn_protection_on and
                                population[victim].ext_is_newborn):
                            
                            # kill the victim
                            del population[victim]
                            
                            # terminate while loop
                            invalid = False
        
        # begin the output string
        ep.ext_out_str += "----------\n"
        
        # if using gradual extinction
        if ep.ext_duration != 1:
            # log extinction information
            ep.ext_out_str += "Gen {} Gradual Extinction - Deaths: {}".format(generation, del_num)
        
        # if using instant extinction
        else:
            # log extinction information
            ep.ext_out_str += "Gen {} Instant Extinction - Deaths: {}".format(generation, del_num)
        
        # if population dynamics are enabled
        if ep.ext_dynamics:
            # if del_num was reduced due to elitism and/or newborn protection restrictions
            if original_num + repopulate > del_num:
                # while the correct amount of members to be repopulated has not been reached
                while (original_num + repopulate != del_num):
                    # increment the amount of members to be repopulated
                    repopulate -= 1
            
            # generate appropriate amount of members to be repopulated
            repop_pop = self._random_repop(repopulate)
            
            # add all newly generated members to the parent population
            population.extend(repop_pop)
            
            # log repopulation information
            ep.ext_out_str += " | Births: {} | Aggregate Deaths: {}".format(repopulate, del_num - repopulate)
        
        # log extinction information
        ep.ext_out_str += "\n"
        
        # if an extinction event is currently in progress
        if ep.ext_in_progress:
            # if this is the final generation for the extinction event, terminate it
            if ep.ext_gens_tracker == ep.ext_duration:
                # turn extinction indicator off
                ep.ext_in_progress = False
                
                # reset extinction event generational tracker
                ep.ext_gens_tracker = 1
                
                # end the output string
                ep.ext_out_str += "----------\n"
        
        # if the extinction event is still in progress
        if ep.ext_in_progress:
            # increment extinction event progress
            ep.ext_gens_tracker += 1
        
        # if the extinction event has run its course
        else:
            # set gradual repopulation variable
            ep.repop_in_progress = True
        
        # return the altered population and the altered evolutionary algorithm parameters
        return population, ret_params
    
    
    # repopulation method, repopulates population back to its original size according to repopulation parameters
    def _repopulate(self, population, generation, ret_params):
        '''
        Parameters
        ----------
        :param population : the sorted parent population
            :type : list
        
        :param generation : the current generation
            :type : int
        
        :param ret_params : the evolutionary algorithm parameters to be altered by repopulation
            :type : list of the form [xo_prob, mut_prob, tourn_size]
        
        Return Value
        ----------
        :return : the altered population, increased in size according to repopulation parameters
            :type : list
        
        :return : the evolutionary algorithm parameters altered by repopulation
            :type : list of the form [xo_prob, mut_prob, tourn_size]
        '''
        
        # if this is the first generation of repopulation
        if ep.repop_gens_tracker == 1:
            # if using gradual repopulation
            if ep.repop_duration != 1:
                # log repopulation information
                ep.ext_out_str += "\n----------\n"
                ep.ext_out_str += "Gradual Repopulation Begins: Generation {}\n".format(generation)
                ep.ext_out_str += "----------\n"
            
            # if using instant repopulation
            else:
                # log repopulation information
                ep.ext_out_str += "\n----------\n"
                ep.ext_out_str += "Instant Repopulation Begins: Generation {}\n".format(generation)
                ep.ext_out_str += "----------\n"
        
        # if this is not the first generation of repopulation
        else:
            # begin the output string
            ep.ext_out_str += "----------\n"
        
        # if using gradual repopulation
        if ep.repop_duration != 1:
            # log repopulation information
            ep.ext_out_str += "Gen {} Gradual Repopulation - ".format(generation)
        
        # if using instant repopulation
        else:
            # log repopulation information
            ep.ext_out_str += "Gen {} Instant Repopulation - ".format(generation)
        
        # set the current population size
        population_size = len(population)
        
        # define child population
        child_population = []
        
        # determine number of members to be repopulated this generation
        repop_amount = self._calculate(ep.ext_original_size, population_size, ep.repop_duration, ep.repop_constant)
        
        # define number of members to repopulate using elite members
        elite_num = 0
        
        # define number of members to repopulate using surviving members
        survive_num = 0
        
        # if using elite repopulation
        if ep.repop_use_elites:
            # determine number of members to repopulate using elite members
            elite_num = int(repop_amount * ep.repop_elite_percent)
            
            # log repopulation information
            ep.ext_out_str += "Elites: {} | ".format(elite_num)
            
            # call the elite repopulation method
            elite_children = self._elite_repop(ep.repop_elite_pop, elite_num)
            
            # add generated elite children to child population
            child_population.extend(elite_children)
        
        # if using survivor repopulation
        if ep.repop_use_survivors:
            # determine number of members to repopulate using surviving members
            survive_num = int(repop_amount * ep.repop_survive_percent)
            
            # log repopulation information
            ep.ext_out_str += "Survivors: {} | ".format(survive_num)
            
            # call the survivor repopulation method
            survive_children = self._survive_repop(population, survive_num)
            
            # add generated survivor children to child population
            child_population.extend(survive_children)
        
        # determine remaining number of members to be generated randomly
        rand_num = repop_amount - (elite_num + survive_num)
        
        # log repopulation information
        ep.ext_out_str += "Random: {} | ".format(rand_num)
        ep.ext_out_str += "Total: {}\n".format(elite_num + survive_num + rand_num)
        
        # randomly generate remaining members
        rand_children = self._random_repop(rand_num)
        
        # add randomly generated members to child population
        child_population.extend(rand_children)
        
        # if the evolutionary algorithm uses the DEAP library
        if ep.ext_uses_DEAP:
            # iterate through the child population and mark them as newborn members
            for member in child_population:
                # indicate that the member is a newborn
                member[1] = True
        
        # if the evolutionary algorithm does not use the DEAP library
        else:
            # iterate through the child population and mark them as newborn members
            for member in child_population:
                # indicate that the member is a newborn
                member.ext_is_newborn = True
        
        # add all newly generated children to the parent population
        population.extend(child_population)
        
        # if alternate parameters are in use
        if ep.alt_params_on:
            # call the method to slowly decrease alternate parameters back to their original values
            ret_params = self._alt_param_decrease(repop_amount, ret_params)
        
        # if repopulation is currently in progress
        if ep.repop_in_progress:
            # if this is the final generation for repopulating, terminate it
            if ep.repop_gens_tracker == ep.repop_duration:
                # turn repopulation off
                ep.repop_in_progress = False
                
                # reset repopulation generational tracker
                ep.repop_gens_tracker = 1
                
                # end the output string
                ep.ext_out_str += "----------\n"
        
        # if repopulation is still in progress
        if ep.repop_in_progress:
            # increment repopulation progress
            ep.repop_gens_tracker += 1
        
        # return the altered population (unsorted and unevaluated) and the altered evolutionary algorithm parameters
        return population, ret_params
    
    
    # generates members based on the current elite members for the repopulate method
    def _elite_repop(self, elite_pop, elite_num):
        '''
        Parameters
        ----------
        :param elite_pop : the elite members of the population
            :type : list
        
        :param elite_num : the number of elite children to generate
            :type : int
        
        Return Value
        ----------
        :return : the sub-population of num children that have been generated from elite members
            :type : list
        '''
        
        # define elite child population
        elite_children = []
        
        # TODO user must finish this code chunk, code instructions are found below
        '''
        1) Generate 'num' amount of children using elite members held in the 'elite_pop' list.
        2) Add generated children to the 'elite_children' list.
        3) Evaluate the fitness of all generated children now held in the 'elite_children' list.
        '''
        
        # return the elite children
        return elite_children
    
    
    # generates members based on the current survivors in the population for the repopulate method
    def _survive_repop(self, survive_pop, survive_num):
        '''
        Parameters
        ----------
        :param survive_pop : the surviving population left over from an extinction event
            :type : list
        
        :param survive_num : the number of survivor children to generate
            :type : int
        
        Return Value
        ----------
        :return : the sub-population of num children that have been generated from surviving members
            :type : list
        '''
        
        # define survivor child population
        survive_children = []
        
        # TODO user must finish this 'if' code block, code instructions are found below
        '''
        1) Generate 'survive_num' amount of children using surviving members held in the 'survive_pop' list.
        2) Add generated children to the 'survive_children' list.
        3) Evaluate the fitness of all generated children now held in the 'survive_children' list.
        '''
        
        # return the survivor children
        return survive_children
    
    
    # generates random members for the repopulate method
    def _random_repop(self, rand_num):
        '''
        Parameters
        ----------
        :param rand_num : the number of random members to generate
            :type : int
        
        Return Value
        ----------
        :return : the sub-population of num members that have been randomly generated
            :type : list
        '''
        
        # define random population
        rand_children = []
        
        # TODO user must finish this code chunk, code instructions are found below
        '''
        1) Generate 'rand_num' amount of children using a user-defined method of randomly generating members.
        2) Add generated children to the 'rand_children' list.
        3) Evaluate the fitness of all generated children now held in the 'rand_children' list.
        '''
        
        # return the random population
        return rand_children
    
    
    # calculates the number of members to kill or repopulate this generation
    def _calculate(self, carrying_capacity, population_size, duration, constant):
        '''
        Parameters
        ----------
        :param carrying_capacity : the maximum population size
            :type : int
        
        :param population_size : the current size of the population
            :type : int
        
        :param duration : the number of generations it will take for extinction or repopulation to complete
            :type : int
        
        :param constant : indicates if a constant value will be used instead of a logarithmic value
            :type : Boolean
        
        Return Value
        ----------
        :return : the number of members to kill or repopulate this generation
            :type : int
        '''
        
        # if this is the first generation of extinction or repopulation
        if (ep.ext_gens_tracker == 1 and ep.ext_in_progress) or (ep.repop_gens_tracker == 1 and ep.repop_in_progress):
            # if calculating for an extinction event
            if ep.ext_gens_tracker == 1 and ep.ext_in_progress:
                # adjust the population size to what it would be after a simulated extinction event
                population_size = population_size - int(ep.ext_percent * population_size)
            
            # cast all relevant arguments to floats
            carrying_capacity = float(carrying_capacity)
            population_size = float(population_size)
            duration = float(duration)
            
            # clear all old birth/death values
            ep.ext_values = ep.ext_values[:]
            
            # calculate total number of members to be killed
            killed = int(ep.ext_percent * ep.ext_original_size)
            
            # calculate total number of members to be repopulated
            repopulated = ep.ext_original_size - population_size
            
            # if the duration lies between one and seven generations or a constant value will be used
            if (1 <= duration <= 7) or constant:
                # calculate all birth/death values for the duration of the extinction event or repopulation
                for gen in range(int(duration)):
                    # if the values are being built for an extinction event
                    if ep.ext_gens_tracker == 1 and ep.ext_in_progress:
                        # calculate the number of members to remove this generation of the duration
                        members = int(killed / duration)
                        
                        # append the number of members to the death values
                        ep.ext_values.append(members)
                    
                    # if the values are being built for repopulation
                    if ep.repop_gens_tracker == 1 and ep.repop_in_progress:
                        # calculate the number of members to add this generation of the duration
                        members = int(repopulated / duration)
                        
                        # append the number of members to the birth values
                        ep.ext_values.append(members)
            
            # if the duration is at least eight generations and logarithmic values will be used
            else:
                # calculate all birth/death values for the duration of the extinction event or repopulation
                for gen in range(int(duration)):
                    # calculate the growth/decay rate of the population
                    rate = (1 / duration) * log((carrying_capacity * population_size), duration) - 0.5 * \
                                              log((population_size / carrying_capacity), duration)
                    
                    # calculate the number of members to add/remove this generation of the duration
                    members = int(rate * (1 - (population_size / carrying_capacity)) * population_size)
                    
                    # append the number of members to the birth/death values
                    ep.ext_values.append(members)
                    
                    # increment the current population size as if the population actually had those members added in
                    population_size += members
            
            # if the values are being built for an extinction event
            if ep.ext_gens_tracker == 1 and ep.ext_in_progress:
                # determine the fraction of individuals that are not zero
                recalculator = len([x for x in ep.ext_values if x is not 0]) / float(len(ep.ext_values))
                
                # ensure the recalculator is not zero
                if recalculator is not 0:
                    # alter each value
                    for idx in range(len(ep.ext_values)):
                        # recalculate the value to be smaller to extend the length of repopulation
                        ep.ext_values[idx] = int(ep.ext_values[idx] * recalculator)
                
                # define loop variable
                invalid = True
                
                # define list index to be altered
                idx = 0
                
                # due to truncation, the death values may be lacking in a few members
                while (invalid):
                    # check if the sum of the death values is equal to the total number of members to be killed
                    if sum(ep.ext_values) != killed:
                        # increment the number of members to be killed at the current index
                        ep.ext_values[idx] += 1
                        
                        # increment the list index
                        idx += 1
                        
                        # ensure the list index has not gone beyond the length of the list
                        if idx >= len(ep.ext_values):
                            # reset list index
                            idx = 0
                    
                    # if all death values are correct
                    else:
                        invalid = False
                
                # for each death value, calculate extra members to be killed for population dynamics
                for amount in ep.ext_values:
                    # calculate additional members to be killed for population dynamics
                    ep.ext_pop_dynamics.append(int(amount * ep.ext_dyn_strength))
                
                # reverse the order of the population dynamics list to accurately reflect a biological ecosystem
                ep.ext_pop_dynamics.reverse()
                
                # if the extinction event gets stronger as it progresses
                if ep.ext_virulence:
                    # reverse the order of the death values list so that the extinction event gets stronger
                    ep.ext_values.reverse()
                    
                    # reverse the order of the population dynamics list to accurately reflect a biological ecosystem
                    ep.ext_pop_dynamics.reverse()
                
                # for each death value, increment it by the amount given by population dynamics
                for idx in range(len(ep.ext_values)):
                    # increment the death value by the amount given by population dynamics
                    ep.ext_values[idx] += ep.ext_pop_dynamics[idx]
            
            # if the values are being built for repopulation
            if ep.repop_gens_tracker == 1 and ep.repop_in_progress:
                # determine the fraction of individuals that are not zero
                recalculator = len([x for x in ep.ext_values if x is not 0]) / float(len(ep.ext_values))
                
                # ensure the recalculator is not zero
                if recalculator is not 0:
                    # alter each value
                    for idx in range(len(ep.ext_values)):
                        # recalculate the value to be smaller to extend the length of repopulation
                        ep.ext_values[idx] = int(ep.ext_values[idx] * recalculator)
                
                # define loop variable
                invalid = True
                
                # define list index to be altered
                idx = 0
                
                # due to truncation, the birth values may be lacking in a few members
                while (invalid):
                    # check if the sum of the birth values is equal to the total number of members to be repopulated
                    if sum(ep.ext_values) != repopulated:
                        # increment the number of members to be repopulated at the current index
                        ep.ext_values[idx] += 1
                        
                        # increment the list index
                        idx += 1
                        
                        # ensure the list index has not gone beyond the length of the list
                        if idx >= len(ep.ext_values):
                            # reset list index
                            idx = 0
                    
                    # if all birth values are correct
                    else:
                        invalid = False
                
                # if repopulation gets stronger as it progresses
                if ep.repop_growth:
                    # reverse the order of the list so that repopulation gets stronger
                    ep.ext_values.reverse()
        
        # return the number of members to kill or repopulate this generation
        return ep.ext_values.pop(0)
    
    
    # increases alternate parameters as the extinction event causes the population to decrease
    def _alt_param_increase(self, del_num, ret_params):
        '''
        Parameters
        ----------
        :param del_num : the amount of members to be killed this generation
            :type : int
        
        :param ret_params : the evolutionary algorithm parameters to be altered by an extinction event
            :type : list of the form [xo_prob, mut_prob, tourn_size]
        
        Return Value
        ----------
        :return : the evolutionary algorithm parameters altered by the extinction event
            :type : list of the form [xo_prob, mut_prob, tourn_size]
        '''
        
        # determine the killed population size through a simulated extinction event
        killed_pop = int(ep.ext_percent * ep.ext_original_size)
        
        # cast the killed population amount to a float
        killed_pop = float(killed_pop)
        
        # if alternate selection parameters are in use and this is the first generation of extinction
        if ep.alt_select_on and (ep.ext_gens_tracker == 1):
            # decrease the tournament size to reduce selection pressure
            ret_params[2] -= ep.alt_tourn_size
        
        # if alternate crossover parameters are in use
        if ep.alt_xo_on:
            # calculate how much to increase crossover operator by for this generation
            alt_xo_inc = ep.alt_xo_prob * (del_num / killed_pop)
            
            # increase crossover rate
            ret_params[0] += alt_xo_inc
        
        # if alternate mutation parameters are in use
        if ep.alt_mut_on:
            # calculate how much to increase mutation operator by for this generation
            alt_mut_inc = ep.alt_mut_prob * (del_num / killed_pop)
            
            # increase mutation rate
            ret_params[1] += alt_mut_inc
        
        # return the altered evolutionary algorithm parameters
        return ret_params
    
    
    # decreases alternate parameters as repopulation causes the population to increase
    def _alt_param_decrease(self, repop_amount, ret_params):
        '''
        Parameters
        ----------
        :param repop_amount : the amount of members to be repopulated this generation
            :type : int
        
        :param ret_params : the evolutionary algorithm parameters to be altered by repopulation
            :type : list of the form [xo_prob, mut_prob, tourn_size]
        
        Return Value
        ----------
        :return : the evolutionary algorithm parameters altered by repopulation
            :type : list of the form [xo_prob, mut_prob, tourn_size]
        '''
        
        # determine the killed population size through a simulated extinction event
        killed_pop = int(ep.ext_percent * ep.ext_original_size) + ep.ext_terminate
        
        # cast the killed population amount to a float
        killed_pop = float(killed_pop)
        
        # if alternate selection parameters are in use and this is the first generation of repopulation
        if ep.alt_select_on and (ep.repop_gens_tracker == 1):
            # increase the tournament size back to its original value to increase selection pressure
            ret_params[2] += ep.alt_tourn_size
        
        # if alternate crossover parameters are in use
        if ep.alt_xo_on:
            # calculate how much to decrease crossover operator by for this generation
            alt_xo_dec = ep.alt_xo_prob * (repop_amount / killed_pop)
            
            # decrease crossover rate
            ret_params[0] -= alt_xo_dec
        
        # if alternate mutation parameters are in use
        if ep.alt_mut_on:
            # calculate how much to decrease mutation operator by for this generation
            alt_mut_dec = ep.alt_mut_prob * (repop_amount / killed_pop)
            
            # decrease mutation rate
            ret_params[1] -= alt_mut_dec
        
        # return the altered evolutionary algorithm parameters
        return ret_params
    
    
    # holds user saved extinction parameters for ext_params.py, may copy/paste the 'if' code block to save a new type
    def _saved_parameters(self, saved_idx):
        '''
        Parameters
        ----------
        :param saved_idx : the index into the set of extinction parameters saved in this method
            :type : int
        '''
        
        # if the testing parameter sets in ext_types.py will be used
        if ep.ext_testing_on:
            # call the given extinction parameter set
            et.saved_params(saved_idx)
        
        # if user generated parameter sets will be used
        else:
            # use default values found in ext_params.py
            if saved_idx == 0:
                pass
            
            # TODO user may copy and paste printed description of parameters
            if saved_idx == 1:
                # TODO user may copy and paste parameters printed to console by enabling 'ext_print_on' in ext_params.py
                pass
    
    
    # takes all extinction parameters used in the current run and stores them in the form of Python code
    def _fetch_params(self):
        # extinction description
        ep.ext_out_str = "\nExtinction Description:\n"
        ep.ext_out_str += "-----------------------\n"
        
        # automatically generate description based on current extinction parameters
        ep.ext_out_str += "'''\n"
        if ep.extinction_on:
            # generate extinction percentage description
            ep.ext_out_str += "Kills {}% of the population\n".format(ep.ext_percent * 100)
            
            # generate extinction duration description
            if ep.ext_duration == 1:
                ep.ext_out_str += "Instant extinction\n"
            else:
                ep.ext_out_str += "Gradual extinction over {} generations\n".format(ep.ext_duration)
            
            if (ep.ext_duration > 7) and not ep.ext_kills_constant:
                ep.ext_out_str += "Logarithmic decrease, "
                
                if ep.ext_virulence:
                    ep.ext_out_str += "gets stronger\n"
                else:
                    ep.ext_out_str += "gets weaker\n"
            elif ep.ext_duration != 1:
                ep.ext_out_str += "Constant decrease\n"
            
            # generate population dynamics description
            if ep.ext_dynamics:
                ep.ext_out_str += "Population dynamics enabled, {}% strength\n".format(ep.ext_dyn_strength * 100)
            else:
                ep.ext_out_str += "Population dynamics disabled\n"
            
            # generate end of extinction description
            if ep.ext_end_extinction:
                ep.ext_out_str += "Extinction possible until {}% of total generations\n" \
                    .format(ep.ext_end_percent * 100)
            else:
                ep.ext_out_str += "Extinction possible for duration of run\n"
            
            # generate safeguard description
            if ep.ext_safeguard == 0:
                ep.ext_out_str += "No safeguard\n"
            else:
                ep.ext_out_str += "Safeguard: {} generations\n".format(ep.ext_safeguard)
            
            # generate elitism description
            if ep.ext_use_elitism:
                ep.ext_out_str += "Elitism: {}% of population\n".format(ep.ext_elite_percent * 100)
            else:
                ep.ext_out_str += "No elitism\n"
            
            # generate anti-elitism description
            if ep.ext_use_anti_elite:
                ep.ext_out_str += "Anti-elitism: {}% of population\n".format(ep.ext_anti_percent * 100)
            else:
                ep.ext_out_str += "No anti-elitism\n"
            
            # generate interval extinction description
            if ep.interval_extinction_on:
                ep.ext_out_str += "Interval Extinction:\n"
                
                if ep.i_ext_occurs:
                    if ep.i_ext_interval == 1:
                        ep.ext_out_str += "    Extinction occurs every generation\n"
                    else:
                        ep.ext_out_str += "    Extinction occurs every {} generations\n".format(ep.i_ext_interval)
                else:
                    if ep.i_ext_interval == 1:
                        ep.ext_out_str += "    Extinction has chance to occur every generation\n"
                    else:
                        ep.ext_out_str += "    Extinction has chance to occur every {} generations\n" \
                            .format(ep.i_ext_interval)
            
            # generate probabilistic extinction description
            if ep.prob_extinction_on:
                ep.ext_out_str += "Probabilistic Extinction:\n"
                ep.ext_out_str += "    Base probability to occur: {}%\n".format(ep.p_ext_base_prob * 100)
                ep.ext_out_str += "    Probability increase: {}%\n".format(ep.p_ext_inc * 100)
            
            # generate fitness extinction description
            if ep.fit_extinction_on:
                ep.ext_out_str += "Fitness Extinction:\n"
                
                if ep.f_ext_use_threshold:
                    ep.ext_out_str += "    {}% of population must pass fitness of {}\n" \
                        .format(ep.f_ext_fit_percent_high * 100, ep.f_ext_threshold_high)
                
                if ep.f_ext_use_stagnation:
                    ep.ext_out_str += "    {}% of population must reach fitness of {}\n" \
                        .format(ep.f_ext_fit_percent_low * 100, ep.f_ext_threshold_low)
            
            # generate repopulation description
            ep.ext_out_str += "Repopulation:\n"
            
            if ep.repop_duration == 1:
                ep.ext_out_str += "    Instant repopulation\n"
            else:
                ep.ext_out_str += "    Gradual repopulation over {} generations\n".format(ep.repop_duration)
            
            if ep.repop_use_elites:
                ep.ext_out_str += "    Elites comprise {}% of repopulated members\n" \
                    .format(ep.repop_elite_percent * 100)
            
            if ep.repop_use_survivors:
                ep.ext_out_str += "    Survivors comprise {}% of repopulated members\n" \
                    .format(ep.repop_survive_percent * 100)
            
            if (ep.repop_duration > 7) and not ep.repop_constant:
                ep.ext_out_str += "    Logarithmic increase, "
                
                if ep.repop_growth:
                    ep.ext_out_str += "gets stronger\n"
                else:
                    ep.ext_out_str += "gets weaker\n"
            elif ep.repop_duration != 1:
                ep.ext_out_str += "    Constant increase\n"
            
            # generate newborn description
            if ep.newborn_params_on:
                ep.ext_out_str += "Newborn Parameters:\n"
                
                ep.ext_out_str += "    Status expires after {} generations\n".format(ep.newborn_expires)
                
                if ep.newborn_protection_on:
                    ep.ext_out_str += "    Newborns protected\n"
                else:
                    ep.ext_out_str += "    Newborns unprotected\n"
            
            # generate alternate parameters description
            if ep.alt_params_on:
                ep.ext_out_str += "Alternate parameters enabled\n"
            else:
                ep.ext_out_str += "Alternate parameters disabled\n"
        
        # if extinction is disabled
        else:
            ep.ext_out_str += "No extinction\n"
        
        ep.ext_out_str += "'''\n"
        ep.ext_out_str += "-----------------------\n"
        
        # print extinction parameters
        ep.ext_out_str += "\n# extinction parameters\n"
        
        ep.ext_out_str += "ep.extinction_on = {}\n".format(ep.extinction_on)
        
        # if extinction is enabled, print all extinction parameters
        if ep.extinction_on:
            # print general extinction parameters
            ep.ext_out_str += "ep.ext_percent = {}\n".format(ep.ext_percent)
            ep.ext_out_str += "ep.ext_duration = {}\n".format(ep.ext_duration)
            
            if (ep.ext_duration > 7) and not ep.ext_kills_constant:
                ep.ext_out_str += "ep.ext_virulence = {}\n".format(ep.ext_virulence)
            else:
                ep.ext_out_str += "ep.ext_kills_constant = {}\n".format(ep.ext_kills_constant)
            
            ep.ext_out_str += "ep.ext_dynamics = {}\n".format(ep.ext_dynamics)
            
            if ep.ext_dynamics:
                ep.ext_out_str += "ep.ext_dyn_strength = {}\n".format(ep.ext_dyn_strength)
            
            ep.ext_out_str += "ep.ext_end_extinction = {}\n".format(ep.ext_end_extinction)
            
            if ep.ext_end_extinction:
                ep.ext_out_str += "ep.ext_end_percent = {}\n".format(ep.ext_end_percent)
            
            ep.ext_out_str += "ep.ext_safeguard = {}\n".format(ep.ext_safeguard)
            
            # print elitism parameters
            ep.ext_out_str += "\n# elitism parameters\n"
            ep.ext_out_str += "ep.ext_use_elitism = {}\n".format(ep.ext_use_elitism)
            
            if ep.ext_use_elitism:
                ep.ext_out_str += "ep.ext_elite_percent = {}\n".format(ep.ext_elite_percent)
                ep.ext_out_str += "ep.ext_elite_target = {}\n".format(ep.ext_elite_target)
            
            # print anti-elitism parameters
            ep.ext_out_str += "\n# anti-elitism parameters\n"
            ep.ext_out_str += "ep.ext_use_anti_elite = {}\n".format(ep.ext_use_anti_elite)
            
            if ep.ext_use_anti_elite:
                ep.ext_out_str += "ep.ext_anti_percent = {}\n".format(ep.ext_anti_percent)
                ep.ext_out_str += "ep.ext_anti_target = {}\n".format(ep.ext_anti_target)
            
            # print extinction type parameters
            ep.ext_out_str += "\n# extinction types\n"
            ep.ext_out_str += "ep.interval_extinction_on = {}\n".format(ep.interval_extinction_on)
            ep.ext_out_str += "ep.prob_extinction_on = {}\n".format(ep.prob_extinction_on)
            ep.ext_out_str += "ep.fit_extinction_on = {}\n".format(ep.fit_extinction_on)
            
            # print interval extinction parameters
            if ep.interval_extinction_on:
                ep.ext_out_str += "\n# interval extinction parameters\n"
                ep.ext_out_str += "ep.i_ext_interval = {}\n".format(ep.i_ext_interval)
                ep.ext_out_str += "ep.i_ext_occurs = {}\n".format(ep.i_ext_occurs)
            
            # print probabilistic extinction parameters
            if ep.prob_extinction_on:
                ep.ext_out_str += "\n# probabilistic extinction parameters\n"
                ep.ext_out_str += "ep.p_ext_base_prob = {}\n".format(ep.p_ext_base_prob)
                ep.ext_out_str += "ep.p_ext_inc = {}\n".format(ep.p_ext_inc)
            
            # print fitness extinction parameters
            if ep.fit_extinction_on:
                ep.ext_out_str += "\n# fitness extinction parameters\n"
                ep.ext_out_str += "ep.f_ext_use_threshold = {}\n".format(ep.f_ext_use_threshold)
                
                if ep.f_ext_use_threshold:
                    ep.ext_out_str += "ep.f_ext_fit_percent = {}\n".format(ep.f_ext_fit_percent_high)
                    ep.ext_out_str += "ep.f_ext_threshold_high = {}\n".format(ep.f_ext_threshold_high)
                
                ep.ext_out_str += "ep.f_ext_use_stagnation = {}\n".format(ep.f_ext_use_stagnation)
                
                if ep.f_ext_use_stagnation:
                    ep.ext_out_str += "ep.f_ext_fit_percent_low = {}\n".format(ep.f_ext_fit_percent_low)
                    ep.ext_out_str += "ep.f_ext_threshold_low = {}\n".format(ep.f_ext_threshold_low)
            
            # print repopulation parameters
            ep.ext_out_str += "\n# repopulation parameters\n"
            ep.ext_out_str += "ep.repop_duration = {}\n".format(ep.repop_duration)
            
            if (ep.repop_duration > 7) and not ep.repop_constant:
                ep.ext_out_str += "ep.repop_growth = {}\n".format(ep.repop_growth)
            else:
                ep.ext_out_str += "ep.repop_constant = {}\n".format(ep.repop_constant)
            
            ep.ext_out_str += "ep.repop_use_elites = {}\n".format(ep.repop_use_elites)
            
            if ep.repop_use_elites:
                ep.ext_out_str += "ep.repop_elite_percent = {}\n".format(ep.repop_elite_percent)
            
            ep.ext_out_str += "ep.repop_use_survivors = {}\n".format(ep.repop_use_survivors)
            
            if ep.repop_use_survivors:
                ep.ext_out_str += "ep.repop_survive_percent = {}\n".format(ep.repop_survive_percent)
            
            # print newborn parameters
            ep.ext_out_str += "\n# newborn parameters\n"
            ep.ext_out_str += "ep.newborn_params_on = {}\n".format(ep.newborn_params_on)
            
            if ep.newborn_params_on:
                ep.ext_out_str += "ep.newborn_expires = {}\n".format(ep.newborn_expires)
                ep.ext_out_str += "ep.newborn_protection_on = {}\n".format(ep.newborn_protection_on)
            
            # print alternate parameters
            ep.ext_out_str += "\n# alternate parameters\n"
            ep.ext_out_str += "ep.alt_params_on = {}\n".format(ep.alt_params_on)
            
            if ep.alt_params_on:
                # print alternate selection parameters
                ep.ext_out_str += "\n# alternate selection parameters\n"
                ep.ext_out_str += "ep.alt_select_on = {}\n".format(ep.alt_select_on)
                
                if ep.alt_select_on:
                    ep.ext_out_str += "ep.alt_tourn_size = {}\n".format(ep.alt_tourn_size)
                
                # print alternate crossover parameters
                ep.ext_out_str += "\n# alternate crossover parameters\n"
                ep.ext_out_str += "ep.alt_xo_on = {}\n".format(ep.alt_xo_on)
                
                if ep.alt_xo_on:
                    ep.ext_out_str += "ep.alt_xo_prob = {}\n".format(ep.alt_xo_prob)
                
                # print alternate mutation parameters
                ep.ext_out_str += "\n# alternate mutation parameters\n"
                ep.ext_out_str += "ep.alt_mut_on = {}\n".format(ep.alt_mut_on)
                
                if ep.alt_mut_on:
                    ep.ext_out_str += "ep.alt_mut_prob = {}\n".format(ep.alt_mut_prob)
        
        ep.ext_out_str += "\n-----------------------\n"
    
    
    # determines if input parameters found in ext_params.py are valid
    def _error_checking(self, population_size, maximum_generations):
        '''
        Parameters
        ----------
        :param population_size : the current size of the population
            :type : int
        
        :param maximum_generations : the total generations the evolutionary algorithm will run for
            :type : int
        '''
        
        # define list to hold errors if they are found
        error_list = []
        
        # ensure percentage of population to kill is in the correct range
        if not 0.00 < ep.ext_percent < 1.00:
            error_list.append("ERROR: 'ext_percent' is out of range! Possible input range: (0.00, 1.00)")
        
        # ensure the duration of extinction is less than the maximum number of generations
        if not 0 < ep.ext_duration < maximum_generations:
            error_list.append("ERROR: 'ext_duration' is out of range! " + \
                              "Possible input range: (0, {})".format(maximum_generations))
        
        # if population dynamics are in use, check for errors
        if ep.ext_dynamics:
            # ensure the population dynamics strength is in the correct range
            if not 0.00 <= ep.ext_dyn_strength <= 1.00:
                error_list.append("ERROR: 'ext_dyn_strength' is out of range! Possible input range: [0.00, 1.00]")
        
        # if ext_end_extinction is in use, check for errors
        if ep.ext_end_extinction:
            # ensure percentage of maximum generations is in the correct range
            if not 0.00 < ep.ext_end_percent <= 1.00:
                error_list.append("ERROR: 'ext_end_percent' is out of range! Possible input range: (0.00, 1.00]")
        
        # if elitism is in use, check for errors
        if ep.ext_use_elitism:
            # ensure percentage of population to consider elite is in the correct range
            if not 0.00 <= ep.ext_elite_percent < 1.00:
                error_list.append("ERROR: 'ext_elite_percent' is out of range! Possible input range: [0.00, 1.00)")
            
            # ensure elitism targeting has been assigned
            if ep.ext_elite_target is None:
                error_list.append("ERROR: 'ext_elite_target' must be assigned a Boolean value!")
        
        # if anti-elitism is in use, check for errors
        if ep.ext_use_anti_elite:
            # ensure percentage of population to consider anti-elite is in the correct range
            if not 0.00 <= ep.ext_anti_percent < 1.00:
                error_list.append("ERROR: 'ext_anti_percent' is out of range! Possible input range: [0.00, 1.00)")
            
            # ensure anti-elitism targeting has been assigned
            if ep.ext_anti_target is None:
                error_list.append("ERROR: 'ext_anti_target' must be assigned a Boolean value!")
        
        # if both elitism and anti-elitism are in use, check for errors
        if ep.ext_use_elitism and ep.ext_use_anti_elite:
            # ensure elitism and anti-elitism are not targeting the same members
            if ep.ext_elite_target == ep.ext_anti_target:
                error_list.append("ERROR: elitism and anti-elitism cannot target same members!\n" + \
                                  "       'ext_elite_target' and 'ext_anti_target' must have opposite truth values")
            
            # ensure elitism and anti-elitism are not overlapping each other
            if not ep.ext_elite_percent + ep.ext_anti_percent <= 1.00:
                error_list.append("ERROR: elitism and anti-elitism percentages cannot overlap!\n" + \
                                  "       'ext_elite_percent' and 'ext_anti_percent' sum cannot exceed 1.00")
        
        # if either probabilistic extinction or fitness extinction is in use, check for errors
        if ep.prob_extinction_on or ep.fit_extinction_on:
            # ensure safeguard is less than the maximum number of generations
            if not 0 <= ep.ext_safeguard < maximum_generations:
                error_list.append("ERROR: 'ext_safeguard' is out of range! " + \
                                  "Possible input range: [0, {})".format(maximum_generations))
        
        # ensure at least one of the extinction types is enabled
        if not (ep.interval_extinction_on or ep.prob_extinction_on or ep.fit_extinction_on):
            error_list.append("ERROR: at least one of the extinction types must be enabled!\n       " + \
                              "'interval_extinction_on', 'prob_extinction_on', or 'fit_extinction_on' must be True")
        
        # if interval extinction is in use, check for errors
        if ep.interval_extinction_on:
            # ensure at least one of the extinction types is enabled if interval extinction won't produce an extinction
            if not ep.i_ext_occurs and not (ep.prob_extinction_on or ep.fit_extinction_on):
                error_list.append("ERROR: at least one of the extinction types must be enabled!\n" + \
                                  "       'prob_extinction_on' or 'fit_extinction_on' must be True")
            
            # ensure the interval is less than the maximum number of generations
            if not 0 < ep.i_ext_interval < maximum_generations:
                error_list.append("ERROR: 'i_ext_interval' is out of range! " + \
                                  "Possible input range: (0, {})".format(maximum_generations))
        
        # if probabilistic extinction is in use, check for errors
        if ep.prob_extinction_on:
            # ensure base extinction probability is in the correct range
            if not 0.00 <= ep.p_ext_base_prob <= 1.00:
                error_list.append("ERROR: 'p_ext_base_prob' is out of range! Possible input range: [0.00, 1.00]")
            
            # ensure the probability increase is in the correct range
            if not 0.00 < ep.p_ext_inc <= 1.00:
                error_list.append("ERROR: 'p_ext_inc' is out of range! Possible input range: (0.00, 1.00]")
        
        # if fitness extinction is in use, check for errors
        if ep.fit_extinction_on:
            # ensure at least one of the fitness extinction types is enabled
            if not (ep.f_ext_use_threshold or ep.f_ext_use_stagnation):
                error_list.append("ERROR: at least one of the fitness extinction types must be enabled!\n" + \
                                  "       'f_ext_use_threshold' or 'f_ext_use_stagnation' must be True")
            
            # if fitness threshold extinction is in use, check for errors
            if ep.f_ext_use_threshold:
                # ensure percentage of population to pass fitness threshold is in the correct range
                if not 0.00 < ep.f_ext_fit_percent_high < 1.00:
                    error_list.append("ERROR: 'f_ext_fit_percent_high' is out of range!" + \
                                      "Possible input range: (0.00, 1.00)")
            
            # if fitness stagnation extinction is in use, check for errors
            if ep.f_ext_use_stagnation:
                # ensure percentage of population to pass fitness threshold is in the correct range
                if not 0.00 < ep.f_ext_fit_percent_low < 1.00:
                    error_list.append("ERROR: 'f_ext_fit_percent_low' is out of range!" + \
                                      "Possible input range: (0.00, 1.00)")
        
        # ensure the repopulation rate is less than the maximum number of generations
        if not 0 < ep.repop_duration < maximum_generations:
            error_list.append("ERROR: 'repop_duration' is out of range! " + \
                              "Possible input range: (0, {})".format(maximum_generations))
        
        # if using elite members for repopulation, check for errors
        if ep.repop_use_elites:
            # ensure elitism is enabled if elite members are needed for repopulation
            if not ep.ext_use_elitism:
                error_list.append("ERROR: elitism must be enabled! 'ext_use_elitism' must be True")
            
            # ensure percentage of population to be repopulated using elite members is in the correct range
            if not 0.00 <= ep.repop_elite_percent <= 1.00:
                error_list.append("ERROR: 'repop_elite_percent' is out of range! Possible input range: [0.00, 1.00]")
        
        # if using surviving members for repopulation, check for errors
        if ep.repop_use_survivors:
            # ensure percentage of population to be repopulated using surviving members is in the correct range
            if not 0.00 <= ep.repop_survive_percent <= 1.00:
                error_list.append("ERROR: 'repop_survive_percent' is out of range! Possible input range: [0.00, 1.00]")
        
        # if using both elite members and surviving members for repopulation, check for errors
        if ep.repop_use_elites and ep.repop_use_survivors:
            # ensure elite member and surviving member percentages are not overlapping each other
            if not ep.repop_elite_percent + ep.repop_survive_percent <= 1.00:
                error_list.append("ERROR: elite and survivor percentages for repopulation cannot overlap!\n" + \
                                  "       'repop_elite_percent' and 'repop_survive_percent' sum cannot exceed 1.00")
        
        # if newborn parameters are in use, check for errors
        if ep.newborn_params_on:
            # ensure number of generations members are considered newborn is less than the maximum number of generations
            if not 0 <= ep.newborn_expires < maximum_generations:
                error_list.append("ERROR: 'newborn_expires' is out of range! " + \
                                  "Possible input range: [0, {})".format(maximum_generations))
        
        # if alternate parameters are in use, check for errors
        if ep.alt_params_on:
            # if alternate selection parameters are in use, check for errors
            if ep.alt_select_on:
                # ensure the alternate tournament size is less than the current population size and not negative
                if not 0 <= ep.alt_tourn_size < population_size:
                    error_list.append("ERROR: 'alt_tourn_size' is out of range! " + \
                                      "Possible input range: [0, {})".format(population_size))
            
            # if alternate crossover parameters are in use, check for errors
            if ep.alt_xo_on:
                # ensure the alternate crossover probability is in the correct range
                if not 0.00 <= ep.alt_xo_prob <= 1.00:
                    error_list.append("ERROR: 'alt_xo_prob' is out of range! Possible input range: [0.00, 1.00]")
            
            # if alternate mutation parameters are in use, check for errors
            if ep.alt_mut_on:
                # ensure the alternate mutation probability is in the correct range
                if not 0.00 <= ep.alt_mut_prob <= 1.00:
                    error_list.append("ERROR: 'alt_mut_prob' is out of range! Possible input range: [0.00, 1.00]")
        
        # if any errors are found, print all errors and exit the program
        if error_list:
            # console spacing
            print("\n_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#\n")
            
            # print all errors found
            for error in error_list:
                # print errors
                print(error)
            
            # print error message
            print("\nFix the above input errors in ext_params.py!")
            
            # console spacing
            print("\n_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#_#\n")
            
            # terminate the program
            sys.exit(1)
